#include "libcDAG.hpp"
#include "config.h"


/** 
* This struct is used to map goal graphs to integers 
* We need the following mapping functions:
*
* int x = get_free_index(): 
* returns an index to a free position in an array of graph* where we can put a graph
*
* free_index(i):
* frees the graph at index i, on a subsequent call to get_free_index() i should be returned
*
* graph* g = get_graph_at(i):
* return the graph at index i
*
* All operations should be (amortized) O(1). This can be done with the following data structure:
*/

struct _cDAG_graph_table_s {
	std::vector<cDAG_Graph> graphs;
	std::vector<unsigned int> free_positions;	
};

struct _cDAG_graph_table_s _cDAG_graph_table; 

struct _cDAG_schedule_table_s {
	std::vector<cDAG_Schedule> schedules;
	std::vector<unsigned int> free_positions;	
};

struct _cDAG_schedule_table_s _cDAG_schedule_table;

struct _cDAG_handle_table_s {
	std::vector<cDAG_Handle> handles;
	std::vector<unsigned int> free_positions;	
};

struct _cDAG_handle_table_s _cDAG_handle_table;



#ifdef __cplusplus
extern "C" {
#endif

/* prototypes */

void FC_FUNC (cdagf_init,cDAGF_INIT)(int* ierr);
void FC_FUNC (cdagf_is_initialized,cDAGF_IS_INITIALIZED)(int* status);
void FC_FUNC (cdagf_finalize,cDAGF_FINALIZE)(int* ierr);
void FC_FUNC (cdagf_is_finalized,cDAGF_IS_FINALIZED)(int* status);
void FC_FUNC (cdagf_my_rank,cDAGF_MY_RANK)(int* rank);
void FC_FUNC (cdagf_group_size,cDAGF_GROUP_SIZE)(int* size);
void FC_FUNC (cdagf_create_graph,cDAGF_CREATE_GRAPH)(int* graph);
void FC_FUNC (cdagf_send,cDAGF_SEND)(int* graph, void* buf, int* count, int* dest, int* memtype, int* vertex_handle);
void FC_FUNC (cdagf_recv,cDAGF_RECV)(int* graph, void* buf, int* count, int* source, int* memtype, int* vertex_handle);
void FC_FUNC (cdagf_requires,cDAGF_REQUIRES)(int* graph, int* vertex_first, int* vertex_after);
void FC_FUNC (cdagf_compile,cDAGF_COMPILE)(int* graph, int* schedule_handle);
void FC_FUNC (cdagf_run,cDAGF_RUN)(int* schedule_handle, int* run_handle);
void FC_FUNC (cdagf_test,cDAGF_TEST)(int* run_handle, int* state);
void FC_FUNC (cdagf_wait,cDAGF_WAIT)(int* run_handle);
void FC_FUNC (cdagf_free_schedule,cDAGF_FREE_SCHEDULE)(int* schedule);

/* wrapper functions */

 


void FC_FUNC (cdagf_init,CDAGF_INIT)(int* ierr) {
	*ierr = cDAG_Init();
}

void FC_FUNC (cdagf_is_initialized,CDAGF_IS_INITIALIZED)(int* status) {
	*status = cDAG_IsInitialized();
}

void FC_FUNC (cdagf_finalize,CDAGF_FINALIZE)(int* ierr) {

	_cDAG_graph_table.graphs.clear();
	*ierr = cDAG_Finalize();

}

void FC_FUNC (cdagf_is_finalized,CDAGF_IS_FINALIZED)(int* status) {
	*status = cDAG_IsFinalized();
}

void FC_FUNC (cdagf_my_rank,CDAGF_MY_RANK)(int* rank) {
	*rank = cDAG_MyRank();
}

void FC_FUNC (cdagf_group_size,CDAGF_GROUP_SIZE)(int* size) {
	*size = cDAG_GroupSize();
}

void FC_FUNC (cdagf_create_graph,CDAGF_CREATE_GRAPH)(int* graph) {

	int pos = -1;
	if (_cDAG_graph_table.free_positions.size() > 0) {
		pos = _cDAG_graph_table.free_positions.back();
		_cDAG_graph_table.free_positions.pop_back();
		_cDAG_graph_table.graphs[pos] = cDAG_Create_graph();
	}
	else {
		pos = _cDAG_graph_table.graphs.size();
		_cDAG_graph_table.graphs.push_back(cDAG_Create_graph());
	}
	
	*graph = pos;

}

void FC_FUNC (cdagf_free_graph,CDAGF_FREE_GRAPH)(int* graph) {

	cDAG_FreeGraph(_cDAG_graph_table.graphs[*graph]);
	_cDAG_graph_table.free_positions.push_back(*graph);

}

void FC_FUNC (cdagf_send,CDAGF_SEND)(int* graph, void* buf, int* count, int* dest, int* memtype, int* vertex_handle) {

	*vertex_handle =  cDAG_Send(_cDAG_graph_table.graphs.at(*graph), buf, *count, *dest, *memtype);

}

void FC_FUNC (cdagf_recv,CDAGF_RECV)(int* graph, void* buf, int* count, int* source, int* memtype, int* vertex_handle) {

	*vertex_handle = cDAG_Recv(_cDAG_graph_table.graphs[*graph], buf, *count, *source, *memtype);

}

void FC_FUNC (cdagf_requires,CDAGF_REQUIRES)(int* graph, int* vertex_first, int* vertex_after) {

	cDAG_Requires(_cDAG_graph_table.graphs[*graph], *vertex_first, *vertex_after);

}

void FC_FUNC (cdagf_compile,CDAGF_COMPILE)(int* graph, int* schedule_handle) {

	int pos = -1;

	if (_cDAG_schedule_table.free_positions.size() > 0) {
		pos = _cDAG_schedule_table.free_positions.back();
		_cDAG_schedule_table.free_positions.pop_back();
		_cDAG_schedule_table.schedules[pos] = cDAG_Compile(_cDAG_graph_table.graphs[*graph]);
	}
	else {
		pos = _cDAG_schedule_table.schedules.size();
		_cDAG_schedule_table.schedules.push_back(cDAG_Compile(_cDAG_graph_table.graphs[*graph]));
	}
	
	*schedule_handle = pos;

}

void FC_FUNC (cdagf_run,CDAGF_RUN)(int* schedule_handle, int* run_handle) {

	int pos = -1;

	if (_cDAG_handle_table.free_positions.size() > 0) {
		pos = _cDAG_handle_table.free_positions.back();
		_cDAG_handle_table.free_positions.pop_back();
		_cDAG_handle_table.handles[pos] = cDAG_Run(_cDAG_schedule_table.schedules[*schedule_handle]);
	}
	else {
		pos = _cDAG_handle_table.handles.size();
		_cDAG_handle_table.handles.push_back(cDAG_Run(_cDAG_schedule_table.schedules[*schedule_handle]));
	}
	
	*run_handle = pos;

}

void FC_FUNC (cdagf_wait,CDAGF_WAIT)(int* run_handle) {

	cDAG_Wait(_cDAG_handle_table.handles[*run_handle]);

	 _cDAG_handle_table.free_positions.push_back(*run_handle);
}

void FC_FUNC (cdagf_free_schedule,CDAGF_FREE_SCHEDULE)(int* schedule) {

	cDAG_FreeSchedule(_cDAG_schedule_table.schedules[*schedule]);
	_cDAG_schedule_table.free_positions.push_back(*schedule);

}


#ifdef __cplusplus
}
#endif

